/****************************************************************************
**
** Copyright (C) 2020 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the examples of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:BSD$
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** BSD License Usage
** Alternatively, you may use this file under the terms of the BSD license
** as follows:
**
** "Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are
** met:
**   * Redistributions of source code must retain the above copyright
**     notice, this list of conditions and the following disclaimer.
**   * Redistributions in binary form must reproduce the above copyright
**     notice, this list of conditions and the following disclaimer in
**     the documentation and/or other materials provided with the
**     distribution.
**   * Neither the name of The Qt Company Ltd nor the names of its
**     contributors may be used to endorse or promote products derived
**     from this software without specific prior written permission.
**
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
** "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
** LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
** A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
** OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
** SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
** LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
** OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
**
** $QT_END_LICENSE$
**
****************************************************************************/

#include <QGuiApplication>
#include <QQmlApplicationEngine>
#include <QtQuick3D/qquick3d.h>
#include <QQmlContext>

#include <QMatrix4x4>
#include <QVector3D>

#include "skingeometry.h"
#include "modelcontroller.h"



bool decomposeMatrix(const QMatrix4x4& matrix, QVector3D* pos, QQuaternion* rotation, QVector3D* scale)
{
    auto m = matrix.data();
    if (pos)
    {
        // Extract the translation.
        pos->setX (m[12]);
        pos->setY (m[13]);
        pos->setZ (m[14]);
    }

    // Nothing left to do.
    if (scale == nullptr && rotation == nullptr)
        return true;

    // Extract the scale.
    // This is simply the length of each axis (row/column) in the matrix.
    const QVector3D xaxis(m[0], m[1], m[2]);
    const float scaleX = xaxis.length();

    const QVector3D yaxis(m[4], m[5], m[6]);
    const float scaleY = yaxis.length();

    const QVector3D zaxis(m[8], m[9], m[10]);
    float scaleZ = zaxis.length();

    // Determine if we have a negative scale (true if determinant is less than zero).
    // In this case, we simply negate a single axis of the scale.
    float det = matrix.determinant();
    if (det < 0)
        scaleZ = -scaleZ;

    if (scale)
    {
        scale->setX (scaleX);
        scale->setY (scaleY);
        scale->setZ (scaleZ);
    }

    // Nothing left to do.
    if (rotation == nullptr)
        return true;

    // Scale too close to zero, can't decompose rotation.
    if (scaleX < 0.00001f || scaleY < 0.00001f || fabs(scaleZ) < 0.00001f)
        return false;

    if(rotation)
    {
        const QMatrix4x4& transform = matrix;
        float matrixData[9] = {transform(0,0), transform(0,1), transform(0,2),
                               transform(1,0), transform(1,1), transform(1,2),
                               transform(2,0), transform(2,1), transform(2,2)};

        QMatrix3x3 rotationMatrix(matrixData);
        *rotation = QQuaternion::fromRotationMatrix(rotationMatrix);
    }

    return true;
}

int main(int argc, char *argv[])
{
    QGuiApplication app(argc, argv);

    QSurfaceFormat::setDefaultFormat(QQuick3D::idealSurfaceFormat());

    QQmlApplicationEngine engine;

    QQmlContext *context = engine.rootContext();

    ModelController controller(engine);
    context->setContextProperty("ModelController", &controller);

    engine.load(QUrl(QStringLiteral("qrc:/main.qml")));
    if (engine.rootObjects().isEmpty())
        return -1;



    QVector3D position(0.0f, 1.0f, 0.0f);
    QQuaternion roation;


    QMatrix4x4 testMatrix;
    testMatrix.translate(position);

    QVector3D startPosition;
    QQuaternion startRotation;
    QVector3D startScale;

    decomposeMatrix(testMatrix, &startPosition, &startRotation, &startScale);
    qDebug() << startPosition << startRotation.toEulerAngles() << startScale;

    testMatrix.rotate(90, QVector3D(1.0f, 0.0f, 0.0f));


    QVector3D endPosition;
    QQuaternion endRotation;
    QVector3D endScale;

    decomposeMatrix(testMatrix, &endPosition, &endRotation, &endScale);
    qDebug() << endPosition << endRotation.toEulerAngles() << endScale;



    {
        QMatrix4x4 newTransform;
        newTransform.scale(QVector3D(2.0f, 2.0f, 2.0f));
        newTransform.rotate(QQuaternion::fromAxisAndAngle(QVector3D(1.0f, 0.0f, 0.0f), -90.0f));
        newTransform.translate(QVector3D(0.0f, 1.0f, 0.0));

        qDebug() << "Test" << newTransform * QVector3D(0.0f, 0.0f, 0.0f);
    }



//    {
//        float xAxis = startPosition.x();
//        float yAxis = startPosition.y();
//        float zAxis = startPosition.z();


//        const float zProxy = zAxis;
//        zAxis = -yAxis;
//        yAxis = zProxy;

//        startPosition = QVector3D(xAxis, yAxis, zAxis);
//    }

//    {
//        QVector3D xAxis;
//        QVector3D yAxis;
//        QVector3D zAxis;
//        startRotation.getAxes(&xAxis, &yAxis, &zAxis);

//        const QVector3D yProxy = yAxis;
//        yAxis = zAxis;
//        zAxis = -yProxy;
//        startRotation = startRotation.fromAxes(xAxis, yAxis, zAxis);
//    }

//    qDebug() << startPosition << startRotation.toEulerAngles() << startScale;


//    QMatrix4x4 newTransform;
//    newTransform.translate(startPosition);
//    newTransform.rotate(startRotation);
//    newTransform.scale(startScale);

//    QVector3D endPosition;
//    QQuaternion endRotation;
//    QVector3D endScale;

//    decomposeMatrix(newTransform, &endPosition, &endRotation, &endScale);
//    qDebug() << endPosition << endRotation.toEulerAngles() << endScale;


    return app.exec();
}
